/* ez-bitset.c: Moderately Scalable set of 32-bit integers.

   Copyright (C) 2019 Michael Zucchi

   This program is free software: you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public License
   as published by the Free Software Foundation, either version 3 of
   the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
   GNU General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this program. If not, see
   <http://www.gnu.org/licenses/>.
*/

#include <stdint.h>
#include <stdlib.h>
#include <string.h>

#include "ez-bitset.h"

/**
 * This implements a bitset using a 3-level sparse table.
 *
 * The final bits are stored as 32-bit integers.
 *
 * This is only moderately cpu and space efficient if large contiguous
 * ranges are set.
 */

/* ********************************************************************** */
/* Private structs */

// L0B + L1B + L2B = 32
#define L0B 9
#define L1B 9
#define L2B 14

#define L0 (1<<L0B)
#define L1 (1<<L1B)
#define L2 ((1<<L2B)/32)

#define L0S (L1B+L2B)
#define L1S (L2B)
#define L2S (5)

#define L0M (L0-1)
#define L1M (L1-1)
#define L2M (L2-1)

#define L3M 31

struct level2 {
	uint32_t bits[L2];
};

struct level1 {
	struct level2 *level2[L1];
};

struct ez_bitset {
	struct level1 *level1[L0];
	uint32_t count;
	// default?
};


// The compiler makes some odd decisions
void ez_bitset_clear(ez_bitset *bs) __attribute__((noinline));
static uint32_t bitset_scan_next(ez_bitset *set, ez_bitset_scan *scan) __attribute__((noinline));

/* ********************************************************************** */

/**
 * Find the next set bit.
 *
 * It first looks for a non-zero bit.
 * Then for a non-zero word in the bit table (at l2).
 * Then for a non-zero table in the l2 table (at l1).
 * Then for a non-zero table in the l1 table (at l0).
 */
static uint32_t bitset_scan_next(ez_bitset *set, ez_bitset_scan *scan) {
	struct level1 *l1 = scan->l1;
	struct level2 *l2 = scan->l2;
	
	while (scan->bits == 0 && scan->id0 < L0) {
		if (l2) {
		retrybit:
			while (scan->id2 < L2) {
				if ((scan->bits = l2->bits[scan->id2++])) {
					scan->offset = (scan->offset & (~L2M << L2S)) | ((scan->id2-1) << L2S);
					goto getbit;
				}
			}
		}

		scan->id2 = 0;

		if (l1) {
		retryl2:
			while (scan->id1 < L1) {
				if ((scan->l2 = l2 = l1->level2[scan->id1++])) {
					scan->offset = (scan->offset & (~L1M<<L1S)) | ((scan->id1-1) << L1S);
					goto retrybit;
				}
			}
		}
		
		scan->id1 = 0;

		while (scan->id0 < L0) {
			if ((scan->l1 = l1 = set->level1[scan->id0++])) {
				scan->offset = (scan->offset & (~L0M << L0S)) | ((scan->id0-1) << L0S);
				goto retryl2;
			}
		}
	}

	if (scan->bits) {
		int index;
	getbit:
		index = ffs(scan->bits) -1 ;
		scan->bits &= ~(1<<index);
		scan->bit = index;
		
		return scan->offset + index;
	}

	return ~0;
}

/* ********************************************************************** */

ez_bitset *ez_bitset_new(void) {
	return calloc(sizeof(ez_bitset), 1);
}

void ez_bitset_free(ez_bitset *bs) {
	ez_bitset_clear(bs);
	free(bs);
}

void ez_bitset_clear(ez_bitset *bs) {
	for (int id0 = 0;id0<L0;id0++) {
		struct level1 *l1 = bs->level1[id0];

		if (l1) {
			for (int id1 = 0; id1 < L1; id1++) {
				struct level2 *l2 = l1->level2[id1];

				free(l2);
			}
			bs->level1[id0] = NULL;
			free(l1);
		}
		
	}
	bs->count = 0;
}

int ez_bitset_get(ez_bitset *bs, uint32_t bit) {
	unsigned int id0 = (bit >> L0S);
	unsigned int id1 = (bit >> L1S) & L1M;
	unsigned int id2 = (bit >> L2S) & L2M;
	unsigned int id3 = 1 << (bit & L3M);

	struct level1 *l1 = bs->level1[id0];

	if (l1) {
		struct level2 *l2 = l1->level2[id1];

		if (l2) {
			return (l2->bits[id2] & id3) != 0;
		}
	}
	
	return 0;	
}

int ez_bitset_set(ez_bitset *bs, uint32_t bit, int value) {
	unsigned int id0 = (bit >> L0S);
	unsigned int id1 = (bit >> L1S) & L1M;
	unsigned int id2 = (bit >> L2S) & L2M;
	unsigned int id3 = 1 << (bit & L3M);
	struct level1 *l1 = bs->level1[id0];
	int old = 0;

	value = value != 0;
	if (value) {
		if (!l1) {
			l1 = calloc(sizeof(*l1), 1);
			bs->level1[id0] = l1;
		}
		struct level2 *l2 = l1->level2[id1];
		if (!l2) {
			l2 = calloc(sizeof(*l2), 1);
			l1->level2[id1] = l2;
		}
		old = (l2->bits[id2] & id3) != 0;
		if (old ^ value) {
			l2->bits[id2] |= id3;
			bs->count += 1;
		}
	} else {
		if (l1) {
			struct level2 *l2 = l1->level2[id1];

			if (l2) {
				old = (l2->bits[id2] & id3) != 0;
				if (old ^ value) {
					l2->bits[id2] &= ~id3;
					bs->count -= 1;
				}
			}
		}
	}
	
	return 0;	
	
}

uint32_t ez_bitset_card(ez_bitset *bs) {
	return bs->count;
}

uint32_t ez_bitset_scan_init(ez_bitset *set, ez_bitset_scan *scan) {
	memset(scan, 0, sizeof(*scan));
	scan->set = set;
	return bitset_scan_next(set, scan);
}

uint32_t ez_bitset_scan_next(ez_bitset_scan *scan) {
	return bitset_scan_next(scan->set, scan);
}

uint32_t ez_bitset_scan_remove(ez_bitset_scan *scan) {
	struct level2 *l2 = scan->l2;

	l2->bits[scan->id2 - 1] &= ~(1<<scan->bit);
	scan->set->count -= 1;
	
	return bitset_scan_next(scan->set, scan);
}
