/* ez-blob-blob.c: Prototype serialiser.

   Copyright (C) 2019 Michael Zucchi

   This program is free software: you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public License
   as published by the Free Software Foundation, either version 3 of
   the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
   GNU General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this program. If not, see
   <http://www.gnu.org/licenses/>.
*/

#include <string.h>
#include <stdlib.h>
#include <assert.h>

#include "ez-list.h"

#include "ez-blob.h"
#include "ez-blob-blob.h"

/*
 This implements a basic 'high performance' serialisation based on the
 descriptors.

 Fields are encoded in the same order as the descriptor with no
 labelling.  Therefore the same descriptor must be used on both ends.
 Fields may be at arbitrary alignments.

 primitives
  encoded in machine order at native size

 strings
  4-byte length followed by characters.  No nul termination.  A length
  of ~0 indicates a null string pointer.

 arrays
  4-byte length followed by data.  ** currently null pointer arrays with zero length are converted to a zero-length array.

 struct
  4-byte length followed by blob-encoded data.

 pointer
  4-byte length followed by blob-encoded data.  A length of ~0
  indicates a null pointer.

*/

static const int primitive_size[EZ_BLOB_FLOAT64+1] = {
	1, 2, 4, 8,
	4, 8
};

size_t ez_blob_size(const ez_blob_desc *desc, const void *p) {
	size_t size = 0;

	for (int i=0,dlen = desc->bd_length;i<dlen;i++) {
		const ez_blob_desc *d = &desc[i+1];

		if (d->bd_type <= EZ_BLOB_FLOAT64) {
			size += primitive_size[d->bd_type];
		} else {
			const void *v = p + d->bd_offset;
			switch (d->bd_type & EZ_BLOB_TYPE) {
			case EZ_BLOB_STRING: {
				char *s = ((char **)v)[0];
				size += s ? 4 + strlen(s) : 4;
				break;
			}
			case EZ_BLOB_BLOB:
				size += 4 + ((const struct ez_blob *)v)->size;
				break;
			case EZ_BLOB_STRUCT: {
				const void *sp = v;
				size += 4 + ez_blob_size(d->bd_table, sp);
				break;
			}
			case EZ_BLOB_POINTER: {
				const void *s = ((const void * const *)v)[0];
				size += s ? 4 + ez_blob_size(d->bd_table, s) : 4;
				break;
			}
			case EZ_BLOB_LIST: {
				ez_list *l = (ez_list *)v;

				size += 4;
				for (ez_node *w = ez_list_head(l), *n = ez_node_succ(w);
				     n;
				     w = n, n=ez_node_succ(n)) {
					size += 4 + ez_blob_size(d->bd_table, w);
				}
				break;
			}
			default:
				break;
			}
		}
	}

	return size;
}

int ez_blob_decode_raw(const ez_blob_desc *desc, const ez_blob *blob, void *p) {
	const char *b = blob->data;
	const char *be = b + blob->size;

	ez_blob_init(desc, p);
	
	for (int i=0,dlen = desc->bd_length;i<dlen;i++) {
		const ez_blob_desc *d = &desc[i+1];
		void *v = (p + d->bd_offset);

		if (d->bd_type <= EZ_BLOB_FLOAT64) {
			int psize = primitive_size[d->bd_type];

			if (b + psize > be)
				goto fail;
			memcpy(v, b, psize);
			b += psize;
		} else if (d->bd_type == EZ_BLOB_LIST) {
			uint32_t count;
			ez_list *l = v;

			if (b + 4 > be)
				goto fail;
			memcpy(&count, b, 4);
			b += 4;

			ez_list_init(l);
			for (int i=0;i<count;i++) {
				uint32_t ss;
				ez_blob sub;
				void *s;

				if (b + 4 > be)
					goto fail;
				memcpy(&ss, b, 4);
				b += 4;

				sub.size = ss;
				sub.data = (void *)b;

				if (b + ss > be)
					goto fail;
				if ((s = calloc(d->bd_table->bd_offset, 1)) == NULL)
					goto fail;
				if (ez_blob_decode_raw(d->bd_table, &sub, s) != 0) {
					free(s);
					goto fail;
				}
				ez_list_addtail(l, s);
				b+=ss;
			}
		} else {
			uint32_t ss;
			ez_blob sub;

			if (b + 4 > be)
				goto fail;
			memcpy(&ss, b, 4);
			b += 4;

			sub.size = ss;
			sub.data = (char *)b;

			switch (d->bd_type & EZ_BLOB_TYPE) {
			case EZ_BLOB_STRING: {
				char *s = NULL;

				if (ss != ~0) {
					if (b + ss > be)
						goto fail;
					if ((s = malloc((size_t)ss + 1)) == NULL)
						goto fail;
					memcpy(s, b, ss);
					s[ss] = 0;
					b += ss;
				}
				((char **)v)[0] = s;
				break;
			}
			case EZ_BLOB_BLOB: {
				struct ez_blob *a = v;
				
				if (b + ss > be)
					goto fail;
				a->size = ss;
				if ((a->data = malloc(ss)) == NULL)
					goto fail;
				memcpy(a->data, b, ss);
				b += ss;
				break;
			}
			case EZ_BLOB_STRUCT:
				if (b + ss > be)
					goto fail;

				if (ez_blob_decode_raw(d->bd_table, &sub, v) != 0)
					goto fail;
				b += ss;
				break;
			case EZ_BLOB_POINTER: {
				void *s = NULL;

				if (ss != ~0) {
					if (b + ss > be)
						goto fail;
					if ((s = calloc(d->bd_table->bd_offset, 1)) == NULL)
						goto fail;
					if (ez_blob_decode_raw(d->bd_table, &sub, s) != 0) {
						free(s);
						goto fail;
					}
					b += ss;
				}
				((void **)v)[0] = s;
				break;
			}
			default:
				break;
			}
		}
	}
	if (b != be)
		goto fail;

	return 0;
 fail:
	ez_blob_free_raw(desc, p);
	return 1;
}

void *ez_blob_decode(const ez_blob_desc *d, const ez_blob *blob) {
	void *p = calloc(d->bd_offset, 1);

	if (ez_blob_decode_raw(d, blob, p) != 0) {
		free(p);
		return NULL;
	}
	
	return p;
}

int ez_blob_encode_raw(const ez_blob_desc *desc, const void *p, ez_blob *blob) {
	char *b = blob->data;
	char *be = b+blob->size;
	
	for (int i=0,dlen = desc->bd_length;i<dlen;i++) {
		const ez_blob_desc *d = &desc[i+1];
		const void *v = (p + d->bd_offset);

		if (d->bd_type <= EZ_BLOB_FLOAT64) {
			int psize = primitive_size[d->bd_type];
			
			assert(b+psize <= be);
			memcpy(b, v, psize);
			b += psize;
		} else if (d->bd_type == EZ_BLOB_LIST) {
			ez_list *l = (void *)v;
			uint32_t ss = ez_list_size(l);

			memcpy(b, &ss, 4);
			b += 4;
			for (ez_node *w = ez_list_head(l), *n = ez_node_succ(w);
			     n;
			     w = n, n=ez_node_succ(n)) {
				ez_blob sub;
				
				ss = ez_blob_size(d->bd_table, w);
				
				assert(b+4+ss <= be);
				memcpy(b, &ss, 4);
				sub.data = b+4;
				sub.size = ss;
				ez_blob_encode_raw(d->bd_table, w, &sub);
				b += 4 + ss;
			}
		} else {
			uint32_t ss;
			struct ez_blob sub;
			
			switch (d->bd_type & EZ_BLOB_TYPE) {
			case EZ_BLOB_STRING: {
				char *s = ((char **)v)[0];
				if (s) {
					ss = strlen(s);
					assert(b+4+ss <= be);
					memcpy(b, &ss, 4);
					memcpy(b+4, s, ss);
					b += 4 + ss;
				} else {
					ss = ~0;
					assert(b+4 <= be);
					memcpy(b, &ss, 4);
					b += 4;
				}
				break;
			}
			case EZ_BLOB_BLOB: {
				const struct ez_blob *a = v;

				ss = a->size;
				assert(b+4+ss <= be);
				memcpy(b, &ss, 4);
				memcpy(b+4, a->data, ss);
				b += 4 + ss;
				break;
			}
			case EZ_BLOB_STRUCT:
				ss = ez_blob_size(d->bd_table, v);
			      
				assert(b+4+ss <= be);
				memcpy(b, &ss, 4);
				sub.data = b+4;
				sub.size = ss;
				ez_blob_encode_raw(d->bd_table, v, &sub);
				b += 4 + ss;
				break;
			case EZ_BLOB_POINTER: {
				const void *s = ((const void **)v)[0];

				if (s) {
					ss = ez_blob_size(d->bd_table, s);
					assert(b+4+ss <= be);
					memcpy(b, &ss, 4);
					sub.data = b+4;
					sub.size = ss;
					ez_blob_encode_raw(d->bd_table, s, &sub);
					b += 4 + ss;
				} else {
					ss = ~0;
					assert(b+4 <= be);
					memcpy(b, &ss, 4);
					b += 4;
				}
				break;
			}
			default:
				break;
			}
		}
	}
	assert(b == be);
	return 0;
}

int ez_blob_encode(const ez_blob_desc *d, const void *p, ez_blob *blob) {
	blob->size = ez_blob_size(d, p);
	blob->data = malloc(blob->size);

	if (blob->data) {
		if (ez_blob_encode_raw(d, p, blob) == 0)
			return 0;
		free(blob->data);
		blob->data = NULL;
	}
	blob->size = 0;
	
	return -1;
}
