/* ez-blob-xdrn.c: XDR-like serialiser using native ordering.

   Copyright (C) 2019 Michael Zucchi

   This program is free software: you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public License
   as published by the Free Software Foundation, either version 3 of
   the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
   GNU General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this program. If not, see
   <http://www.gnu.org/licenses/>.
*/

/*
  Implements an XDR serialiser.

  Each end must have identical descriptor tables.

  STRING or POINTER files which may be NULL must be marked as
  EZ_BLOB_NULLABLE.  They are implemented as optional fields.

*/

#include <stdlib.h>
#include <string.h>
#include <stdint.h>

#include "ez-list.h"

#include "ez-blob.h"
#include "ez-blob-xdrn.h"
#include "ez-blob-io.h"

static size_t roundup(size_t v) __attribute__ ((always_inline));
static __inline__ size_t roundup(size_t v) {
	return (v+3) & ~3;
}

static void xdrn_encode_raw(struct ez_blobio *io, const ez_blob_desc *desc, const void *p) {
	for (int i=0,dlen=desc->bd_length;i<dlen;i++) {
		const ez_blob_desc *d = &desc[i+1];
		const void *v = (p + d->bd_offset);

		switch (d->bd_type) {
		case EZ_BLOB_INT8:
			blobio_write32(io, *(uint8_t *)v);
			break;
		case EZ_BLOB_INT16:
			blobio_write32(io, *(uint16_t *)v);
			break;
		case EZ_BLOB_INT32:
			blobio_write32(io, *(uint32_t *)v);
			break;
		case EZ_BLOB_INT64:
			blobio_write64(io, *(uint64_t *)v);
			break;
		case EZ_BLOB_FLOAT32:
			blobio_writef(io, *(float *)v);
			break;
		case EZ_BLOB_FLOAT64:
			blobio_writed(io, *(double *)v);
			break;
		case EZ_BLOB_STRING | EZ_BLOB_NULLABLE:
			if (!*(char **)v) {
				blobio_write32(io, 0);
				break;
			}
			blobio_write32(io, 1);
			// falls through
		case EZ_BLOB_STRING: {
			int32_t size = strlen(*(char **)v);

			blobio_write32(io, size);
			blobio_write(io, *(char **)v, size);
			blobio_write_align(io, 4);
			break;
		}
		case EZ_BLOB_BLOB: {
			uint32_t size = ((ez_blob *)v)->size;

			blobio_write32(io, size);
			blobio_write(io, ((ez_blob *)v)->data, size);
			blobio_write_align(io, 4);
			break;
		}
		case EZ_BLOB_STRUCT:
			xdrn_encode_raw(io, d->bd_table, v);
			break;
		case EZ_BLOB_POINTER | EZ_BLOB_NULLABLE:
			if (!*(void **)v) {
				blobio_write32(io, 0);
				break;
			}
			blobio_write32(io, 1);
			// falls through
		case EZ_BLOB_POINTER:
			xdrn_encode_raw(io, d->bd_table, *(void **)v);
			break;
		case EZ_BLOB_LIST:
			blobio_write32(io, ez_list_size((ez_list *)v));
			for (ez_node *w = ez_list_head((ez_list *)v), *n = ez_node_succ(w);
			     n;
			     w = n, n=ez_node_succ(n)) {
				xdrn_encode_raw(io, d->bd_table, w);
			}
			break;
		default:
			io->error = 1;
			return;
		}
	}
}

size_t ez_xdrn_size(const ez_blob_desc *d, const void *p) {
	struct ez_blobio io = {
		.mode = BLOBIO_WRITE_SIZE
	};
	xdrn_encode_raw(&io, d, p);

	return io.index;
}

int ez_xdrn_encode_raw(const ez_blob_desc *d, const void *p, ez_blob *blob) {
	struct ez_blobio io = {
		.data = blob->data,
		.size = blob->size,
		.mode = BLOBIO_WRITE_FIXED
	};

	xdrn_encode_raw(&io, d, p);

	return io.error;
}

int ez_xdrn_encode(const ez_blob_desc *d, const void *p, ez_blob *blob) {
	struct ez_blobio io = {
		.mode = BLOBIO_WRITE_ALLOC
	};

	xdrn_encode_raw(&io, d, p);

	if (!io.error) {
		blob->size = io.index;
		blob->data = io.data;
		return 0;
	}

	free(blob->data);
	
	return -1;
}

static void xdrn_decode_raw(struct ez_blobio *io, const ez_blob_desc *desc, void *p) {
	ez_blob_init(desc, p);
	
	for (int i=0,dlen=desc->bd_length;i<dlen;i++) {
		const ez_blob_desc *d = &desc[i+1];
		void *v = (p + d->bd_offset);

		switch (d->bd_type) {
		case EZ_BLOB_INT8:
			*(uint8_t *)v = blobio_readi32(io);
			break;
		case EZ_BLOB_INT16:
			*(uint16_t *)v = blobio_readi32(io);
			break;
		case EZ_BLOB_INT32:
			*(uint32_t *)v = blobio_readi32(io);
			break;
		case EZ_BLOB_INT64:
			*(uint64_t *)v = blobio_readi64(io);
			break;
		case EZ_BLOB_FLOAT32:
			*(float *)v = blobio_readf(io);
			break;
		case EZ_BLOB_FLOAT64:
			*(double *)v = blobio_readd(io);
			break;
		case EZ_BLOB_STRING | EZ_BLOB_NULLABLE:
			if (blobio_readi32(io) == 0)
				break;
			// falls through
		case EZ_BLOB_STRING: {
			uint32_t size = blobio_readi32(io);
			void *src = blobio_take(io, size);

			if (src) {
				char *mem = malloc(size+1);

				if (!mem)
					goto error;

				memcpy(mem, src, size);
				mem[size] = 0;
				*(char **)v = mem;
				blobio_read_align(io, 4);
			}
			break;
		}
		case EZ_BLOB_BLOB: {
			uint32_t size = blobio_readi32(io);
			void *src = blobio_take(io, size);

			if (src) {
				char *mem = malloc(size);

				if (!mem)
					goto error;

				memcpy(mem, src, size);
				((ez_blob *)v)->data = mem;
				((ez_blob *)v)->size = size;
				blobio_read_align(io, 4);
			}
			break;
		}
		case EZ_BLOB_STRUCT:
			xdrn_decode_raw(io, d->bd_table, v);
			break;
		case EZ_BLOB_POINTER | EZ_BLOB_NULLABLE:
			if (blobio_readi32(io) == 0)
				break;
		case EZ_BLOB_POINTER: {
			void *mem = malloc(d->bd_table->bd_offset);

			if (!mem)
				goto error;

			*(void **)v = mem;
			xdrn_decode_raw(io, d->bd_table, mem);
			break;
		}
		case EZ_BLOB_LIST: {
			uint32_t count = blobio_readi32(io);

			for (int j=0;j<count;j++) {
				void *node = malloc(d->bd_table->bd_offset);

				if (!node)
					goto error;

				xdrn_decode_raw(io, d->bd_table, node);
				ez_list_addtail((ez_list *)v, node);
			}
			break;
		}
		default:
			goto error;
		}
	}
	return;
 error:
	io->index = io->size;
	io->error = 1;
}

int ez_xdrn_decode_raw(const ez_blob_desc *desc, const ez_blob *blob, void *p) {
	struct ez_blobio io = {
		.size = blob->size,
		.data = blob->data,
		.mode = BLOBIO_READ
	};

	xdrn_decode_raw(&io, desc, p);

	if (io.error) {
		ez_blob_free_raw(desc, p);
		return -1;
	}

	return 0;
}

void *ez_xdrn_decode(const ez_blob_desc *desc, const ez_blob *blob) {
	void *p = malloc(desc->bd_offset);

	if (p && ez_xdrn_decode_raw(desc, blob, p) != 0) {
		free(p);
		p = NULL;
	}

	return p;
}
